# -*- encoding: utf8 -*-


import sqlite3

conn = None


def get_connection(sqlite_db_path) -> sqlite3.Connection:
    global conn
    if conn is None:
        conn = sqlite3.connect(sqlite_db_path)
        # conn.execute("PRAGMA journal_mode=WAL;")
    return conn




# -*- encoding:utf8 -*-

import functools
import re
import urllib
import urllib.error
from abc import ABC, abstractmethod
from typing import List

import ipaddress
import pendulum
from datetime import datetime
import sqlite3
import logging

from ipwhois import IPWhois, exceptions, ASNRegistryError
from importlib_resources import read_text

from .log_parser import LogEntry, DATETIME_FORMAT_PATTERN, int_to_ip
from . import log_parser
from . import db_connection

LOGGER = logging.getLogger(__name__)


def init_database(sqlite_db_path: str):
    sql_script = read_text("find2deny", "log-data.sql")
    conn = db_connection.get_connection(sqlite_db_path)
    try:
        with conn:
            conn.executescript(sql_script)
            conn.commit()
    except sqlite3.OperationalError as ex:
        LOGGER.error(ex)
        raise JudgmentException(f"Cannot init database in sqlite file {sqlite_db_path}", errors=ex)
    pass


def is_ready_blocked(log_entry: LogEntry, conn: sqlite3.Connection) -> (bool, str):
    @functools.lru_cache(maxsize=2024)
    def __cached_query(ip: int):
        try:
            with conn:
                c = conn.cursor()
                c.execute("SELECT COUNT(*), cause_of_block cause_of_block FROM block_network WHERE ip = ?", (ip,))
                row = c.fetchone()
                ip_count = row[0]
                cause = row[1]
                return (ip_count == 1), cause
        except sqlite3.OperationalError as ex:
            raise AccessDatabaseError("", errors=ex)

    return __cached_query(log_entry.ip)


def update_deny(ip_network: str, log_entry: LogEntry, judge: str, cause_of_block: str, sqlite_db_path: str):
    insert_cmd = "INSERT OR IGNORE INTO block_network (ip, ip_network, block_since, judge, cause_of_block) VALUES (?, ?, ?, ?, ?)"
    try:
        with db_connection.get_connection(sqlite_db_path) as conn:
            conn.execute(insert_cmd, (log_entry.ip, ip_network, local_datetime(), judge, cause_of_block))
    except sqlite3.OperationalError as ex:
        raise AccessDatabaseError(sqlite_db_path, errors=ex)
    LOGGER.info("(%s) add %s to blocked network", log_entry.ip_str, ip_network)
    pass


class AbstractIpJudgment(ABC):

    @abstractmethod
    def should_deny(self, log_entry: LogEntry, entry_count: int = 0) -> (bool, str):
        """
            check if the given ip should be blocked
        :param log_entry: in integer
        :param entry_count: how many entries are processed since python script starts
        :return: (True, cause) if the ip should be blocked, (False,cause) if the firewall should allow ip
        """
        pass


class ChainedIpJudgment(AbstractIpJudgment):

    def __init__(self, conn: sqlite3.Connection, chains: List[AbstractIpJudgment], white_list: List[str] = None):
        self.__white_list: List[str] = white_list or []
        self.__judgment = chains
        # self.__log_db_path = log_db_path
        self.conn = conn  # db_connection.get_connection(self.__log_db_path)

    def should_deny(self, log_entry: LogEntry, entry_count=0) -> bool:
        white_list_fn = make_ip_check_fn(log_entry.ip_str)
        white_lister = next((item for item in self.__white_list if white_list_fn(item)), None)
        if white_lister:
            return False, "White listed by {}".format(white_lister)
        else:
            deny, cause = is_ready_blocked(log_entry, self.conn)
            if deny:
                return deny, cause
            for judgment in self.__judgment:
                deny, cause = judgment.should_deny(log_entry, entry_count)
                if deny:
                    ip_network = lookup_ip(log_entry.ip)
                    update_deny(ip_network, log_entry, judgment.__class__.__name__, cause, self.conn)
                    return True, cause
            return False, None


def make_ip_check_fn(ip):
    return lambda white_list: \
        white_list == ip or is_ip_matched_network(ip, white_list) or is_ip_in_white_listed_network(ip, white_list)


def is_ip_in_white_listed_network(ip, white_list_network):
    try:
        return ipaddress.ip_address(ip) in ipaddress.ip_network(white_list_network)
    except Exception:
        return False


def is_ip_matched_network(ip, white_list_ip_reg):
    pattern = re.compile(white_list_ip_reg)
    return pattern.match(ip)


class PathBasedIpJudgment(AbstractIpJudgment):
    """

    """

    def __init__(self, bot_path: set = None):
        self._bot_path = bot_path if bot_path is not None else {}
        pass

    def should_deny(self, log_entry: LogEntry, entry_count: int = 0) -> (bool, str):
        try:
            request_path = log_entry.request.split(" ")[1]
            request_resource = next((r for r in self._bot_path if request_path.startswith(r)), None)
            blocked = request_resource is not None
            cause = None
            if blocked:
                cause = "{} tried to access {} which matches non-existing resource {}".format(
                    log_entry.ip_str, request_path, request_resource)
                LOGGER.info(cause)
            return blocked, cause
        except IndexError:
            if log_entry.request == "-":
                return False, None
            else:
                cause = "{}-s  request >>{}<< is not conform to HTTP".format(
                    log_entry.ip_str, log_entry.request)
                LOGGER.info(cause)
                return True, cause

    def __str__(self):
        return "PathBasedIpJudgment/bot_path:{}".format(self._bot_path)


class TimeBasedIpJudgment(AbstractIpJudgment):

    def __init__(self, sqlite_db_path: str, allow_access: int = 10, interval_second: int = 10):
        """
        :param path: path to a SQLite Database file
        :param allow_access: number of access in a given time interval (next parameter)
        :param interval_second: time interval in seconds
        """
        self.allow_access = allow_access
        self.interval = interval_second
        self._sqlite_db_path = sqlite_db_path
        self.conn = db_connection.get_connection(self._sqlite_db_path)
        self.count_ip = 0

    def __del__(self):
        pass

    def should_deny(self, log_entry: LogEntry, entry_count: int = 0) -> (bool, str):
        """

        :param log_entry:
        :param entry_count:
        :return:
        """
        self.count_ip = entry_count
        if not self._ready_processed(log_entry):
            return self._make_block_ip_decision(log_entry)
        else:
            return self._lookup_decision_cache(log_entry)

    def _ready_processed(self, log_entry: LogEntry) -> bool:
        sql_cmd = "SELECT count(*) FROM processed_log_ip WHERE ip = ? AND line = ? AND log_file = ?"
        try:
            with self.conn as conn:
                conn.row_factory = sqlite3.Row
                c = conn.cursor()
                c.execute(sql_cmd, (log_entry.ip, log_entry.line, log_entry.log_file))
                row = c.fetchone()
            if not row or row is None:
                return False
            else:
                LOGGER.debug("found %d processed ip in database log for entry %s ", row[0], log_entry)
                ip_count = row[0]
                return ip_count == 1
        except sqlite3.OperationalError:
            LOGGER.warning("Cannot make connection to database file %s", self._sqlite_db_path)
            return False

    def _make_block_ip_decision(self, log_entry: LogEntry) -> (bool, str):
        ip_int = log_entry.ip
        try:
            with self.conn as conn:
                conn.row_factory = sqlite3.Row
                c = conn.cursor()
                c.execute("INSERT INTO processed_log_ip (ip, line, log_file) VALUES (?, ?, ?)",
                          (log_entry.ip, log_entry.line, log_entry.log_file))
                c.execute("""SELECT ip, first_access, last_access, access_count 
                                    FROM log_ip 
                                    WHERE ip = ? 
                                    ORDER BY first_access ASC""",
                          (ip_int,))
                row = c.fetchone()
        except sqlite3.OperationalError as ex:
            raise AccessDatabaseError(self._sqlite_db_path, errors=ex)

        if row is None:
            LOGGER.debug("IP %s not found in log_ip", log_entry.ip_str)
            self._add_log_entry(log_entry)
            LOGGER.info("    [%d] added %s to log_ip", self.count_ip, log_entry.ip_str)
            return False, None
        else:
            first_access = datetime.strptime(row['first_access'], DATETIME_FORMAT_PATTERN)
            LOGGER.debug("log time: %s  first access: %s", log_entry.time, first_access)
            delay = (log_entry.time - first_access).total_seconds()
            access_count = row['access_count'] + 1
            LOGGER.debug("%s accessed %s %s times in %d seconds", log_entry.ip_str, log_entry.request, access_count,
                         delay)
            limit_rate = self.allow_access / self.interval
            if delay >= 0:
                try:
                    access_rate = access_count / delay
                    if access_rate >= limit_rate and delay <= self.interval:
                        cause = "{} accessed server {}-times in {} secs which is too much for rate {} accesses / {}".format(
                            log_entry.ip_str, access_count, delay, self.allow_access, self.interval)
                        self._update_deny(log_entry, access_count)
                        LOGGER.info(cause)
                        return True, cause
                    else:
                        self._update_access(log_entry, access_count)
                        return False, None
                    pass
                except ZeroDivisionError as ex:
                    if access_count > self.allow_access:
                        self._update_deny(log_entry, access_count)
                        cause = "{} accessed server {} in less than 0 secs which is too much for rate {} accesses / {}".format(
                            log_entry.ip_str, access_count, delay, self.allow_access, self.interval)
                        LOGGER.info(cause)
                        return True, cause
                    else:
                        self._update_access(log_entry, access_count)
                        return False, None
            else:
                row_content = self._row2str(row)
                LOGGER.warning(
                    "entry {} is not sorted by modified date: first log: {} last access: {} time-diff {}".
                    format(log_entry, row_content, log_entry.time, delay))
                return False, None

    def _lookup_decision_cache(self, log_entry: LogEntry) -> (bool, str):
        try:
            with self.conn as conn:
                for row in conn.execute("SELECT count(*) as count_ip, cause_of_block FROM block_network WHERE ip = ?",
                                        (log_entry.ip,)):
                    pass
            count = row['count_ip']
            cause = row['cause_of_block'] or None
            return (count == 1), cause
        except sqlite3.OperationalError as ex:
            raise AccessDatabaseError(self._sqlite_db_path, errors=ex)

    def _add_log_entry(self, log_entry: LogEntry):
        time_iso = log_entry['time'].strftime(DATETIME_FORMAT_PATTERN)
        sql_cmd = """INSERT INTO log_ip (ip, first_access, last_access, access_count) 
                                         VALUES (?, ?, ?, ?)"""
        try:
            with self.conn as conn:
                conn.execute(sql_cmd, (log_entry.ip,
                                       time_iso,
                                       time_iso,
                                       1)
                             )
        except sqlite3.OperationalError:
            LOGGER.warning("Cannot insert new log to log_ip")
        pass

    def _update_deny(self, log_entry: LogEntry, access_count: int):
        """
        :param log_entry:
        :param access_count:
        :return:
        """
        ip_network = lookup_ip(log_entry.ip)
        update_cmd = """UPDATE log_ip SET 
            ip_network = ?,
            last_access = ?,
            access_count = ?,
            status = 1
            WHERE ip = ?
        """
        try:
            with self.conn as conn:
                conn.execute(update_cmd, (ip_network, log_entry.iso_time, access_count, log_entry.ip))
        except sqlite3.OperationalError:
            LOGGER.warning("Cannot update log_ip")
        pass

    def _update_access(self, log_entry: LogEntry, access_count: int):
        """

        :param log_entry:
        :param access_count:
        :return:
        """
        try:
            update_cmd = "UPDATE log_ip SET last_access = ?,  access_count = ? WHERE ip = ?"
            with self.conn as conn:
                conn.execute(update_cmd, (local_datetime(), access_count, log_entry.ip))
            LOGGER.debug("update access_count of %s to %s", log_entry.ip_str, access_count)
        except sqlite3.OperationalError:
            print("Cannot update log_ip")
        pass

    def __str__(self):
        return "TimeBasedIpBlocker/database:{}".format(self._sqlite_db_path)

    @staticmethod
    def _row2str(row):
        return "ip:{} first_access:{} last_access:{} access_count:{}".format(
            int_to_ip(row["ip"]),
            row["first_access"],
            row["last_access"],
            row["access_count"]
        )


def local_datetime() -> str:
    return pendulum.now().strftime(DATETIME_FORMAT_PATTERN)


def lookup_ip(ip: str or int) -> str:
    str_ip = ip if isinstance(ip, str) else log_parser.int_to_ip(ip)
    return __lookup_ip(str_ip)


@functools.lru_cache(maxsize=10240)
def __lookup_ip(normed_ip: str) -> str:
    try:
        who = IPWhois(normed_ip).lookup_rdap()
        cidr = who["network"]["cidr"]
        asn_cidr = who["asn_cidr"]
        return cidr if cidr == asn_cidr else (cidr + ', ' + asn_cidr)
    except (urllib.error.HTTPError, exceptions.HTTPLookupError, exceptions.IPDefinedError, ASNRegistryError) as ex:
        LOGGER.warning("IP Lookup for %s fail", normed_ip)
        LOGGER.warning("return ip instead of network")
        LOGGER.debug(ex)
        return normed_ip


class UserAgentBasedIpJudgment(AbstractIpJudgment):
    """
        deny an ip in a log entry if a log entry has a User-Agent String containing one of given
        substring. The list of substrings is given by initialize this class
    """

    def __init__(self, blacklist_agent: List[str]):
        self._blacklist_agent = blacklist_agent

    def should_deny(self, log_entry: LogEntry, entry_count: int = 0) -> (bool, str):
        ua = log_entry.user_agent
        for bl in self._blacklist_agent:
            if ua.find(bl) >= 0:
                cleaned_ua = ua.replace("\n", ' ')
                return True, f"{cleaned_ua} contains {bl}"
        return False, None

    pass


class JudgmentException(Exception):
    def __init__(self, message, errors=None):
        self.message = message
        self.errors = errors
        super(JudgmentException, self).__init__(message)


class AccessDatabaseError(JudgmentException):
    def __init__(self, db_path: str = "", errors=None):
        msg = "Access to Sqlite Db caused error; Diagnose: use `find2deny-init-db {}' to create a Database.".format(db_path)
        super(AccessDatabaseError, self).__init__(msg, errors)


from abc import ABC, abstractmethod

from .log_parser import LogEntry
import re
import logging


class AbstractIpBlockExecution(ABC):
    @abstractmethod
    def begin_execute(self):
        pass

    @abstractmethod
    def block(self,log: LogEntry):
        pass

    @abstractmethod
    def end_execute(self):
        pass


class FileBasedUWFBlock(AbstractIpBlockExecution):
    """
        write generated uwf firewall rules into a file.
        block also all IPs in same Netmask of the IP
    """

    def __init__(self, destinate_path: str, insert_line: str or int = 5):
        self.__destinate_path = destinate_path
        self.__blocked_item = []
        self.__insert_line = insert_line

    def begin_execute(self):
        self.__blocked_item = []
        pass

    def block(self, log: LogEntry, cause: str = None):
        cause = cause.replace('\n', ' ') if cause is not None else ' '
        # Block network of IP
        for n in re.split(",\\s*", log.network):
            ufw_block_cmd = f"ufw insert {self.__insert_line} deny from {n.strip()} to any # {cause}"
            logging.debug("use command %s", ufw_block_cmd)
            self.__blocked_item.append(f"{ufw_block_cmd}\n")
        # block IP
        ufw_block_cmd = f"ufw insert 4 deny from {log.ip_str} to any # {cause}"
        logging.debug("use command %s", ufw_block_cmd)
        self.__blocked_item.append(f"{ufw_block_cmd}\n")
        pass

    def end_execute(self):
        with open(self.__destinate_path, 'w') as f:
            f.write("#!/bin/bash\n")
            f.writelines(self.__blocked_item)
        logging.info("Block %d IPs", len(self.__blocked_item))
        pass


import sys
from os import path
import argparse
import logging
import sqlite3
import glob
import hashlib

from typing import List, Dict
from pprint import pprint, pformat

from .config_parser import ParserConfigException, \
    VERBOSITY, LOG_LEVELS, CONF_FILE, \
    LOG_FILES, LOG_PATTERN, DATABASE_PATH, \
    WHITE_LIST, \
    JUDGMENT, RULES, \
    BOT_REQUEST, MAX_REQUEST, INTERVAL_SECONDS, \
    BLACKLIST_AGENT, \
    EXECUTION, SCRIPT, \
    parse_config_file, INSERT_LINE

from . import log_parser
from . import judgment
from . import execution
from . import db_connection


LOGGER = logging.getLogger(__name__)

# work-flow
# 1. suche alle IP in Logfile nach Merkmale eines Angriff
# 2. generiert UFW Befehlen aus der IPs im Schritte 1
# 3. gebe die Befehlen aus, oder sonstiges weiter Verarbeitung

# CLI options:

# [judgment]

# []

_parser = None


def main():
    global _parser
    argv = sys.argv
    parser = argparse.ArgumentParser()
    parser.add_argument(f"{CONF_FILE}", type=str, nargs="?",
                        help="Configuration file, configuration must be given by a configuration file "
                        f"(the positional argument {CONF_FILE}). The Debug level can be changed by "
                        f"optional argument --{VERBOSITY}.")

    parser.add_argument("-v", f"--{VERBOSITY}",
                        choices=LOG_LEVELS,
                        help="how much information is printed out during processing log files")
    _parser = parser
    cli_arg = vars( parser.parse_args(argv[1:]) )
    verbosity = cli_arg[VERBOSITY]
    if verbosity is not None:
        apply_log_config(verbosity)
    file_based_config = parse_config_file(cli_arg[CONF_FILE])
    if verbosity is not None:
        file_based_config[VERBOSITY] = verbosity
    if logging.getLogger("root").isEnabledFor(logging.DEBUG): LOGGER.debug(pformat(file_based_config))
    validate_config(file_based_config)
    try:
        apply_log_config(file_based_config[VERBOSITY])
        return analyse_log_files(file_based_config)
    except judgment.JudgmentException as ex:
        print(ex, file=sys.stderr)


def apply_log_config(verbosity: str):
    log_level = logging.getLevelName(verbosity)
    logging.getLogger().setLevel(level=log_level)
    logging.basicConfig(
        format='%(relativeCreated)d %(levelname)s %(module)s : %(message)s',
        datefmt='%H:%M:%S',
    )
    LOGGER.info("Verbosity: %s %d", verbosity, log_level)


def validate_config(config: Dict):
    if LOG_FILES not in config:
        raise ParserConfigException("Log files are not configured")


# TODO: test this method
def analyse_log_files(config: Dict):
    log_files = expand_log_files(config[LOG_FILES])
    conn = db_connection.get_connection(config[DATABASE_PATH])
    log_files = filter_processed_files(log_files, conn)
    LOGGER.info("Analyse %d file(s)", len(log_files))
    judge = construct_judgment(config)
    output_shell_script = config[EXECUTION][0][RULES][SCRIPT]
    insert_line = config[EXECUTION][0][RULES][INSERT_LINE]
    executor = execution.FileBasedUWFBlock(output_shell_script, insert_line=insert_line)
    executor.begin_execute()
    log_pattern = config[LOG_PATTERN]
    i = 0
    try:
        for file_path in log_files:
            LOGGER.info("Analyse file %s", file_path)
            update_processed_file(file_path[0], file_path[1], conn)
            logs = log_parser.parse_log_file(file_path[1], log_pattern)
            for log in logs:
                i += 1
                LOGGER.debug("                       [%d] Process `%s'", i, log)
                blocked, cause = judgment.is_ready_blocked(log, conn)
                if blocked:
                    LOGGER.info("IP %s is ready blocked", log.ip_str)
                else:
                    deny, cause = judge.should_deny(log, i)
                    if deny:
                        network = judgment.lookup_ip(log.ip_str)
                        log.network = network
                        executor.block(log, cause)
    except KeyboardInterrupt:  # Will not work with python -m cProfile
        LOGGER.warning("Stop processing log files")
        LOGGER.info("current log files: {}".format(file_path))
        LOGGER.info("Write ready processed log entries to files")
        executor.end_execute()
        try:
            conn.close()
        except Exception as ex:
            LOGGER.warning("Cannot close Sql DbConnection {}", ex)
        return 1
    executor.end_execute()
    return 0


def expand_log_files(config_log_file: List[str]) -> List[str]:
    log_files = []
    for p in config_log_file:
        expand_path = glob.glob(p)
        LOGGER.debug("expand glob '%s' to %s", p, expand_path)
        if len(expand_path) == 0:
            LOGGER.warn("Glob path '%s' cannot be expanded to any real path", p)
        log_files = log_files + expand_path
    return log_files


def apache_access_log_file_chronological_decode(file_name):
    base_name = path.basename(file_name).split('.')
    return -int(base_name[2]) if len(base_name) > 2 else 0


def filter_processed_files(log_files:List[str], conn: sqlite3.Connection, key=apache_access_log_file_chronological_decode)->List[str]:
    processed_files = []
    try:
        with conn:
            conn.row_factory = sqlite3.Row
            c = conn.cursor()
            for row in c.execute("SELECT content_hash, path FROM processed_log_file"):
                processed_files.append(row[0])
    except sqlite3.OperationalError:
        LOGGER.warning(
            "Cannot read table processed_files in database so use all expanded files")
    log_files_hash = map(lambda f: (content_hash(f), f), log_files)
    effective_log_files = sorted(filter(lambda hf: hf[0] not in processed_files, log_files_hash), key=lambda hf: key(hf[1]))
    return list(effective_log_files)


def update_processed_file(hash_content, file_path, conn: sqlite3.Connection):    # TODO: UNIT TEST
    if file_path.endswith("gz"):
        try:
            with conn:
                c = conn.cursor()
                c.execute("""
                    INSERT OR REPLACE INTO processed_log_file (content_hash, path) 
                    VALUES (?,
                        COALESCE((SELECT path FROM processed_log_file WHERE content_hash = ? AND path = ?), ?)
                    )
                    """,
                    (hash_content,hash_content,file_path, file_path)
                )
        except sqlite3.OperationalError as ex:
            LOGGER.warning("Cannot update table processed_files %s", ex)
    else:
        LOGGER.info("File %s is not compressed, so not mark it as processed", file_path)


def content_hash(file_path):
    h = hashlib.sha256()
    with open(file_path, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b""):
            h.update(chunk)
    return h.hexdigest()


def construct_judgment(config) -> judgment.AbstractIpJudgment:
    judgments_chain = config[JUDGMENT] if JUDGMENT in config else []
    if len(judgments_chain) < 1:
        _parser.error(f"At least one entry in {JUDGMENT} must be configured")
    list_of_judgments = []
    for judge in judgments_chain:
        list_of_judgments += [judgment_by_name(judge, config)]
    LOGGER.info("Use %d judgments", len(list_of_judgments))
    return judgment.ChainedIpJudgment(db_connection.get_connection(config[DATABASE_PATH]), list_of_judgments, config[WHITE_LIST])


def judgment_by_name(judge, config):
    name = judge['name']
    rules = judge[RULES]
    if name == "path-based-judgment":
        bot_request_path = rules[BOT_REQUEST] if BOT_REQUEST in rules else []
        if len(bot_request_path) > 0:
            return judgment.PathBasedIpJudgment(bot_request_path)
        else:
            _parser.error(f"At least one path in {BOT_REQUEST} must be configured if Judgment {name} is used")
    elif name == "time-based-judgment":
        if DATABASE_PATH in config:
            database_path = config[DATABASE_PATH]
            conn = db_connection.get_connection(database_path)
            max_request = rules[MAX_REQUEST] if MAX_REQUEST in rules else 500
            interval = rules[INTERVAL_SECONDS] if INTERVAL_SECONDS in rules else 60
            return judgment.TimeBasedIpJudgment(conn, max_request, interval)
        else:
            _parser.error(f"A SQLite database ({DATABASE_PATH}) must be configured in global section if Judgment {name} is used")
    elif name == "user-agent-based-judgment":
        blacklist_agent = rules[BLACKLIST_AGENT] if BLACKLIST_AGENT in rules else []
        return judgment.UserAgentBasedIpJudgment(blacklist_agent)
    else:
        raise ParserConfigException(f"Unknown judgment {name}")


#############################################################################
#############################################################################
#############################################################################
def init_db():
    global _parser
    argv = sys.argv
    parser = argparse.ArgumentParser()
    parser.add_argument("db_path",
                        help="Path to an Sqlite Database")
    _parser = parser
    cli = parser.parse_args(argv[1:])
    db_path = cli.db_path
    judgment.init_database(db_path)
    pass